"""Script for testing the performance of pickling/unpickling.

This will pickle/unpickle several real world-representative objects a few
thousand times. The methodology below was chosen for was chosen to be similar
to real-world scenarios which operate on single objects at a time. Note that if
we did something like

    pickle.dumps([dict(some_dict) for _ in range(10000)])

this isn't equivalent to dumping the dict 10000 times: pickle uses a
highly-efficient encoding for the n-1 following copies.
"""

import datetime
import random
import sys

import pyperf
from memray_helper import get_tracker

IS_PYPY = pyperf.python_implementation() == "pypy"

__author__ = "collinwinter@google.com (Collin Winter)"


DICT = {
    "ads_flags": 0,
    "age": 18,
    "birthday": datetime.date(1980, 5, 7),
    "bulletin_count": 0,
    "comment_count": 0,
    "country": "BR",
    "encrypted_id": "G9urXXAJwjE",
    "favorite_count": 9,
    "first_name": "",
    "flags": 412317970704,
    "friend_count": 0,
    "gender": "m",
    "gender_for_display": "Male",
    "id": 302935349,
    "is_custom_profile_icon": 0,
    "last_name": "",
    "locale_preference": "pt_BR",
    "member": 0,
    "tags": ["a", "b", "c", "d", "e", "f", "g"],
    "profile_foo_id": 827119638,
    "secure_encrypted_id": "Z_xxx2dYx3t4YAdnmfgyKw",
    "session_number": 2,
    "signup_id": "201-19225-223",
    "status": "A",
    "theme": 1,
    "time_created": 1225237014,
    "time_updated": 1233134493,
    "unread_message_count": 0,
    "user_group": "0",
    "username": "collinwinter",
    "play_count": 9,
    "view_count": 7,
    "zip": "",
}

TUPLE = (
    [
        265867233,
        265868503,
        265252341,
        265243910,
        265879514,
        266219766,
        266021701,
        265843726,
        265592821,
        265246784,
        265853180,
        45526486,
        265463699,
        265848143,
        265863062,
        265392591,
        265877490,
        265823665,
        265828884,
        265753032,
    ],
    60,
)


def mutate_dict(orig_dict, random_source):
    new_dict = dict(orig_dict)
    for key, value in new_dict.items():
        rand_val = random_source.random() * sys.maxsize
        if isinstance(key, (int, bytes, str)):
            new_dict[key] = type(key)(rand_val)
    return new_dict


random_source = random.Random(5)  # Fixed seed.
DICT_GROUP = [mutate_dict(DICT, random_source) for _ in range(3)]


def bench_pickle(loops, pickle, options):
    range_it = range(loops)

    # micro-optimization: use fast local variables
    dumps = pickle.dumps
    objs = (DICT, TUPLE, DICT_GROUP)
    protocol = options.protocol

    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range_it:
            for obj in objs:
                # 20 dumps
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)
                dumps(obj, protocol)

        return pyperf.perf_counter() - t0


def bench_unpickle(loops, pickle, options):
    pickled_dict = pickle.dumps(DICT, options.protocol)
    pickled_tuple = pickle.dumps(TUPLE, options.protocol)
    pickled_dict_group = pickle.dumps(DICT_GROUP, options.protocol)
    range_it = range(loops)

    # micro-optimization: use fast local variables
    loads = pickle.loads
    objs = (pickled_dict, pickled_tuple, pickled_dict_group)

    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range_it:
            for obj in objs:
                # 20 loads dict
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)
                loads(obj)

        return pyperf.perf_counter() - t0


LIST = [[list(range(10)), list(range(10))] for _ in range(10)]


def bench_pickle_list(loops, pickle, options):
    range_it = range(loops)
    # micro-optimization: use fast local variables
    dumps = pickle.dumps
    obj = LIST
    protocol = options.protocol

    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range_it:
            # 10 dumps list
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)
            dumps(obj, protocol)

        return pyperf.perf_counter() - t0


def bench_unpickle_list(loops, pickle, options):
    pickled_list = pickle.dumps(LIST, options.protocol)
    range_it = range(loops)

    # micro-optimization: use fast local variables
    loads = pickle.loads

    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range_it:
            # 10 loads list
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)
            loads(pickled_list)

        return pyperf.perf_counter() - t0


MICRO_DICT = dict((key, dict.fromkeys(range(10))) for key in range(100))


def bench_pickle_dict(loops, pickle, options):
    range_it = range(loops)
    # micro-optimization: use fast local variables
    protocol = options.protocol
    obj = MICRO_DICT

    with get_tracker():
        t0 = pyperf.perf_counter()
        for _ in range_it:
            # 5 dumps dict
            pickle.dumps(obj, protocol)
            pickle.dumps(obj, protocol)
            pickle.dumps(obj, protocol)
            pickle.dumps(obj, protocol)
            pickle.dumps(obj, protocol)

        return pyperf.perf_counter() - t0


BENCHMARKS = {
    # 20 inner-loops: don't count the 3 pickled objects
    "pickle": (bench_pickle, 20),
    # 20 inner-loops: don't count the 3 unpickled objects
    "unpickle": (bench_unpickle, 20),
    "pickle_list": (bench_pickle_list, 10),
    "unpickle_list": (bench_unpickle_list, 10),
    "pickle_dict": (bench_pickle_dict, 5),
}


def is_accelerated_module(module):
    return getattr(module.Pickler, "__module__", "<jython>") != "pickle"


def add_cmdline_args(cmd, args):
    if args.pure_python:
        cmd.append("--pure-python")
    cmd.extend(("--protocol", str(args.protocol)))
    cmd.append(args.benchmark)


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "Test the performance of pickling."

    parser = runner.argparser
    parser.add_argument(
        "--pure-python", action="store_true", help="Use the C version of pickle."
    )
    parser.add_argument(
        "--protocol",
        action="store",
        default=None,
        type=int,
        help="Which protocol to use (default: highest protocol).",
    )
    benchmarks = sorted(BENCHMARKS)
    parser.add_argument("benchmark", choices=benchmarks)

    options = runner.parse_args()
    benchmark, inner_loops = BENCHMARKS[options.benchmark]

    name = options.benchmark
    if options.pure_python:
        name += "_pure_python"

    if not (options.pure_python or IS_PYPY):
        # C accelerators are enabled by default on 3.x
        import pickle

        if not is_accelerated_module(pickle):
            raise RuntimeError("Missing C accelerators for pickle")
    else:
        sys.modules["_pickle"] = None
        import pickle

        if is_accelerated_module(pickle):
            raise RuntimeError("Unexpected C accelerators for pickle")

    if options.protocol is None:
        options.protocol = pickle.HIGHEST_PROTOCOL
    runner.metadata["pickle_protocol"] = str(options.protocol)
    runner.metadata["pickle_module"] = pickle.__name__

    runner.bench_time_func(name, benchmark, pickle, options, inner_loops=inner_loops)



from pathlib import Path
from typing import Any
from typing import Dict

import matplotlib
import matplotlib.patches as mpatches
import numpy as np
from matplotlib import pyplot as plt

matplotlib.use("agg")


def get_data(result) -> Dict[str, Any]:
    results = {}

    for case in result:
        for benchmark in case["benchmarks"]:
            if "metadata" in benchmark:
                name = benchmark["metadata"]["name"]
            else:
                name = case["metadata"]["name"]
            data = []
            for run in benchmark["runs"]:
                data.extend(run.get("values", []))
            results[name] = np.array(data, dtype=np.float64)

    return results


def remove_outliers(values, m=2):
    return values[abs(values - np.mean(values)) < m * np.std(values)]


def plot_diff_pair(ax, ref, head, names, outlier_rejection=True):
    master_data = []
    all_data = []
    means = []
    for name in names:
        ref_values = ref[name]
        head_values = head[name]
        if outlier_rejection:
            ref_values = remove_outliers(ref_values)
            head_values = remove_outliers(head_values)
        means.append(np.mean(head_values))
        values = np.outer(ref_values, 1.0 / head_values).flatten()
        values.sort()
        idx = np.round(np.linspace(0, len(values) - 1, 100)).astype(int)
        all_data.append(values[idx])
        master_data.extend(values)

    all_data.append(master_data)
    violin = ax.violinplot(
        all_data,
        vert=False,
        showmeans=True,
        showmedians=False,
        showextrema=False,
        widths=1.0,
        quantiles=[[0.1, 0.9]] * len(all_data),
    )
    violin["cquantiles"].set_linestyle(":")
    for pc, mean in zip(violin["bodies"], means):
        if mean < 0.1:
            pc.set_facecolor("#ff0000")
        else:
            pc.set_facecolor("#0000ff")


def formatter(val, pos):
    return f"{val:.02f}x"


def plot_diff(ref, head, output_filename: Path, title: str) -> None:
    head_data = get_data(head.data)
    ref_data = get_data(ref.data)

    names = sorted(head_data.keys(), reverse=True)
    fig, axs = plt.subplots(figsize=(8, 2 + len(names) * 0.3), layout="constrained")
    plt.axvline(1.0)
    plot_diff_pair(axs, ref_data, head_data, names)
    names.append("ALL")
    axs.set_yticks(np.arange(len(names)) + 1, names)
    axs.set_ylim(0, len(names) + 1)
    axs.tick_params(axis="x", bottom=True, top=True, labelbottom=True, labeltop=True)
    axs.xaxis.set_major_formatter(formatter)
    axs.grid()
    axs.set_title(title)

    # Add legend for colors (red -> microbenchmark, blue -> macrobenchmark)

    axs.legend(
        handles=[
            mpatches.Patch(color="#ff0000", label="Microbenchmark"),
            mpatches.Patch(color="#0000ff", label="Macrobenchmark"),
        ],
        bbox_to_anchor=(0, 1.05),
    )

    plt.savefig(output_filename)
    plt.close()



from _memray.hooks cimport Allocator
from libc.stdint cimport uint64_t
from libc.stdint cimport uintptr_t
from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector


cdef extern from "hooks.h" namespace "memray::hooks":
    cdef enum Allocator:
        MALLOC "memray::hooks::Allocator::MALLOC"
        CALLOC "memray::hooks::Allocator::CALLOC"
        REALLOC "memray::hooks::Allocator::REALLOC"
        VALLOC "memray::hooks::Allocator::VALLOC"
        ALIGNED_ALLOC "memray::hooks::Allocator::ALIGNED_ALLOC"
        POSIX_MEMALIGN "memray::hooks::Allocator::POSIX_MEMALIGN"
        MEMALIGN "memray::hooks::Allocator::MEMALIGN"
        PVALLOC "memray::hooks::Allocator::PVALLOC"
        FREE "memray::hooks::Allocator::FREE"
        PYMALLOC_MALLOC "memray::hooks::Allocator::PYMALLOC_MALLOC"
        PYMALLOC_CALLOC "memray::hooks::Allocator::PYMALLOC_CALLOC"
        PYMALLOC_REALLOC "memray::hooks::Allocator::PYMALLOC_REALLOC"
        PYMALLOC_FREE "memray::hooks::Allocator::PYMALLOC_FREE"

cdef extern from "records.h" namespace "memray::tracking_api":
   ctypedef unsigned long thread_id_t
   ctypedef size_t frame_id_t
   ctypedef size_t code_object_id_t
   ctypedef long long millis_t

   struct CodeObjectInfo:
       string function_name
       string filename
       string linetable
       int firstlineno

   struct AllocationRecord:
       uintptr_t address
       size_t size
       Allocator allocator
       frame_id_t native_frame_id

   struct Frame:
       code_object_id_t code_object_id
       int instruction_offset
       bool is_entry_frame

   struct FramePush:
       Frame frame

   struct FramePop:
       size_t count

   struct ThreadRecord:
       const char* name

   struct Segment:
       uintptr_t vaddr
       size_t memsz

   struct ImageSegments:
       string filename
       uintptr_t addr
       vector[Segment] segments

   struct UnresolvedNativeFrame:
       uintptr_t ip
       frame_id_t index

   struct TrackerStats:
       size_t n_allocations
       size_t n_frames
       long long start_time
       long long end_time

   cdef enum FileFormat:
       ALL_ALLOCATIONS 'memray::tracking_api::FileFormat::ALL_ALLOCATIONS'
       AGGREGATED_ALLOCATIONS 'memray::tracking_api::FileFormat::AGGREGATED_ALLOCATIONS'

   struct HeaderRecord:
       int version
       bool native_traces
       FileFormat file_format
       TrackerStats stats
       string command_line
       int pid
       size_t main_tid
       size_t skipped_frames_on_main_tid
       int python_allocator
       bool trace_python_allocators
       bool track_object_lifetimes
       string libdest
       vector[string] site_packages
       vector[string] sys_path

   cdef cppclass Allocation:
       thread_id_t tid
       uintptr_t address
       size_t size
       Allocator allocator
       frame_id_t native_frame_id
       size_t frame_index
       size_t native_segment_generation
       size_t n_allocations

       object toPythonObject()

   cdef cppclass AggregatedAllocation:
       thread_id_t tid
       Allocator allocator
       frame_id_t native_frame_id
       size_t frame_index
       size_t native_segment_generation

       size_t n_allocations_in_high_water_mark
       size_t n_allocations_leaked
       size_t bytes_in_high_water_mark
       size_t bytes_leaked

       Allocation contributionToHighWaterMark()
       Allocation contributionToLeaks()

   struct MemoryRecord:
       uint64_t ms_since_epoch
       size_t rss

   struct MemorySnapshot:
       uint64_t ms_since_epoch
       size_t rss
       size_t heap

   cdef cppclass TrackedObject:
       thread_id_t tid
       uintptr_t address
       bool is_created
       frame_id_t native_frame_id
       size_t frame_index
       size_t native_segment_generation

       object toPythonObject()

   cdef cppclass ObjectRecord:
       uintptr_t address
       bool is_created
       frame_id_t native_frame_id

cdef extern from "<optional>":
   # Cython doesn't have libcpp.optional yet, so just declare this opaquely.
   cdef cppclass optional_location_id_t "std::optional<memray::api::location_id_t>":
       pass



import math

import pytest
from exercise_3.lru_cache import Algorithms
from exercise_3.lru_cache import compare_counts_different_factorials
from exercise_3.lru_cache import generate_factorial_plus_last_digit

# Memory tests


@pytest.mark.limit_memory("200 MB")
def test_lru_cache():
    compare_counts_different_factorials()


# Correctness tests


@pytest.mark.parametrize("value", [1, 4, 10])
def test_algorithms_0(value):
    a = Algorithms(0)  # This is equivalent to the standard factorial function
    assert a.factorial_plus(value) == math.factorial(value)


def test_algorithms_5():
    a = Algorithms(5)

    # 3 * (2 * (1 + 5) + 5) + 5
    assert a.factorial_plus(3) == 56

    # 4 * (3 * (2 * (1 + 5) + 5) + 5) + 5
    assert a.factorial_plus(4) == 229


def test_generate_factorial_plus_last_digit_0():
    values = list(generate_factorial_plus_last_digit(1, 6))
    assert values == [1, 1, 2, 6, 4, 0]  # last digits of 1 1 2 6 24 120, i.e. 0! to 5!


def test_generate_factorial_plus_0_last_digit():
    values = list(generate_factorial_plus_last_digit(5, 1))
    assert values == [
        1,
        2,
        3,
        4,
        5,
    ]  # last digits of the first fac_plus_n factorial of 0, which is always n+1


def test_generate_factorial_plus():
    values = list(generate_factorial_plus_last_digit(3, 5))
    expected = (
        [1, 1, 2, 6, 4]
        + [2, 2, 5, 6, 5]  # last digits of the first fac_plus_0 factorial of n, i.e. n!
        + [3, 3, 8, 6, 6]  # fac_plus_1 values are [2, 2, 5, 16, 65]
    )  # fac_plus_1 values are [3, 3, 8, 26, 106]
    assert (
        values == expected
    )  # last digits of the first fac_plus(n) factorial of 0, which is always n+1



"""
Convert Docutils' documentation from reStructuredText to <format>.
"""

import contextlib
from pathlib import Path

import docutils
from docutils import core
import pyperf
from memray_helper import get_tracker

from docutils.utils.math.math2html import Trace

Trace.show = lambda message, channel: ...  # don't print to console

DOC_ROOT = (Path(__file__).parent / "docutils_data" / "docs").resolve()


def build_html(doc_root):
    elapsed = 0.0
    for file in doc_root.rglob("*.txt"):
        file_contents = file.read_text(encoding="utf-8")
        t0 = pyperf.perf_counter()
        with get_tracker():
            with contextlib.suppress(docutils.ApplicationError):
                core.publish_string(
                    source=file_contents,
                    reader_name="standalone",
                    parser_name="restructuredtext",
                    writer_name="html5",
                    settings_overrides={
                        "input_encoding": "unicode",
                        "output_encoding": "unicode",
                        "report_level": 5,
                    },
                )
        elapsed += pyperf.perf_counter() - t0
    return elapsed


def bench_docutils(loops, doc_root):
    runs_total = 0
    for _ in range(loops):
        runs_total += build_html(doc_root)
    return runs_total


def add_cmdline_args(cmd, args):
    cmd.append("--doc_root=%s" % args.doc_root)


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "Render documentation with Docutils"
    runner.argparser.add_argument("--doc_root", default=DOC_ROOT)

    args = runner.parse_args()
    runner.bench_time_func("docutils", bench_docutils, Path(args.doc_root))



import threading
from pathlib import Path

from memray import AllocatorType
from memray import FileReader
from memray import Tracker
from memray._test import MemoryAllocator
from memray._test import set_thread_name
from tests.utils import filter_relevant_allocations
from tests.utils import skip_if_macos

HERE = Path(__file__).parent
TEST_MULTITHREADED_EXTENSION = HERE / "multithreaded_extension"


def allocating_function(allocator, flag_event, wait_event):
    allocator.valloc(1234)
    allocator.free()
    flag_event.set()
    wait_event.wait()
    allocator.valloc(1234)
    allocator.free()


def test_thread_allocations_after_tracker_is_deactivated(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    wait_event = threading.Event()
    flag_event = threading.Event()
    allocator = MemoryAllocator()

    # WHEN
    with Tracker(output):
        t = threading.Thread(
            target=allocating_function, args=(allocator, flag_event, wait_event)
        )
        t.start()
        flag_event.wait()

    # Keep allocating in the same thread while the tracker is not active
    wait_event.set()
    t.join()

    # THEN
    relevant_records = list(
        filter_relevant_allocations(FileReader(output).get_allocation_records())
    )
    assert len(relevant_records) == 2

    vallocs = [
        record
        for record in relevant_records
        if record.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 1
    (valloc,) = vallocs
    assert valloc.size == 1234

    frees = [
        record for record in relevant_records if record.allocator == AllocatorType.FREE
    ]
    assert len(frees) == 1


@skip_if_macos
def test_thread_name(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    allocator = MemoryAllocator()

    def allocating_function():
        set_thread_name("my thread name")
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output):
        t = threading.Thread(target=allocating_function)
        t.start()
        t.join()

    # THEN
    relevant_records = list(
        filter_relevant_allocations(FileReader(output).get_allocation_records())
    )
    assert len(relevant_records) == 2

    vallocs = [
        record
        for record in relevant_records
        if record.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 1
    (valloc,) = vallocs
    assert valloc.size == 1234
    assert "my thread name" == valloc.thread_name


def test_setting_python_thread_name(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    allocator = MemoryAllocator()
    name_set_inside_thread = threading.Event()
    name_set_outside_thread = threading.Event()
    prctl_rc = -1

    def allocating_function():
        allocator.valloc(1234)
        allocator.free()

        threading.current_thread().name = "set inside thread"
        allocator.valloc(1234)
        allocator.free()

        name_set_inside_thread.set()
        name_set_outside_thread.wait()
        allocator.valloc(1234)
        allocator.free()

        nonlocal prctl_rc
        prctl_rc = set_thread_name("set by prctl")
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with Tracker(output):
        t = threading.Thread(target=allocating_function, name="set before start")
        t.start()
        name_set_inside_thread.wait()
        t.name = "set outside running thread"
        name_set_outside_thread.set()
        t.join()

    # THEN
    expected_names = [
        "set before start",
        "set inside thread",
        "set outside running thread",
        "set by prctl" if prctl_rc == 0 else "set outside running thread",
    ]
    names = [
        rec.thread_name
        for rec in FileReader(output).get_allocation_records()
        if rec.allocator == AllocatorType.VALLOC
    ]
    assert names == expected_names



import pytest
from exercise_1.fibonacci import generate_fibonacci_hash

# Memory tests


@pytest.mark.limit_memory("100 KB")
def test_fibonacci():
    LENGTH_OF_SEQUENCE_1 = 33333  # pylint: disable=invalid-name
    LENGTH_OF_SEQUENCE_2 = 30000  # pylint: disable=invalid-name
    LENGTH_OF_SEQUENCE_3 = 34567  # pylint: disable=invalid-name

    generate_fibonacci_hash(
        LENGTH_OF_SEQUENCE_1, LENGTH_OF_SEQUENCE_2, LENGTH_OF_SEQUENCE_3
    )


# Correctness tests


def test_fibonacci_empty():
    h = generate_fibonacci_hash(0, 0, 0)
    assert h == 0


@pytest.mark.parametrize(
    ("length", "expected"),
    [
        (1, 1),
        (2, 2),  # 1 + 1
        (6, 20),  # 1 + 1 + 2 + 3 + 5 + 8
    ],
)
def test_fibonacci_short(length, expected):
    h = generate_fibonacci_hash(0, 0, length)
    assert h == expected


@pytest.mark.parametrize(
    ("length", "expected"),
    [
        (1, 1),
        (2, 2),  # 1 + 1
        (6, 20),  # 1 + 1 + 2 + 3 + 5 + 8
    ],
)
def test_fibonacci_multiple(length, expected):
    h = generate_fibonacci_hash(length, length, length)
    assert h == expected * 3


def test_hash_modulo_10000():
    # 1 + 1 + 2 +3 + 5 + 8 + 13 + 21 + 34 + 55 + 89 + 144 + 233 + 377 + 610
    # + 987 + 1597 + 2584 + 4181 == 10945
    h = generate_fibonacci_hash(0, 0, 19)
    assert h == 945  # 10945 % 10000



import pytest

from memray.reporters.frame_tools import is_cpython_internal
from memray.reporters.frame_tools import is_frame_from_import_system
from memray.reporters.frame_tools import is_frame_interesting


class TestFrameFiltering:
    @pytest.mark.parametrize(
        "frame, expected",
        [
            [
                (
                    "_PyEval_EvalFrameDefault",
                    "/src/python/python3.8/Python/ceval.c",
                    100,
                ),
                True,
            ],
            [
                ("_PyEvalSomeFunc", "/src/python/python3.8/Python/ceval.c", 100),
                True,
            ],
            [("VectorCall", "/src/python/python3.8/Python/ceval.c", 100), True],
            [("proxy_call", "/src/python/python3.8/Python/ceval.c", 100), True],
            [
                (
                    "function_code_fastcall",
                    "/src/python/python3.8/Modules/gcmodule.c",
                    100,
                ),
                True,
            ],
            # Python 3.14 tail call interpreter LLVM-generated functions
            [
                (
                    "_TAIL_CALL_INSTRUMENTED_CALL.llvm.10282351651392433962",
                    "/src/python/python3.14/Python/ceval.c",
                    100,
                ),
                True,
            ],
            [
                (
                    "_TAIL_CALL_POP_TOP.llvm.123456789",
                    "/src/python/python3.14/Python/ceval.c",
                    50,
                ),
                True,
            ],
            [("somefunc", "myapp.py", 100), False],
            [("function_code_fastcall", "myapp.py", 100), False],
            # _TAIL_CALL_ pattern should not match without .llvm.
            [
                ("_TAIL_CALL_SOMETHING", "/src/python/python3.14/Python/ceval.c", 100),
                False,
            ],
            # _TAIL_CALL_ pattern should not match in non-CPython files
            [("_TAIL_CALL_SOMETHING.llvm.123", "myapp.py", 100), False],
        ],
    )
    def test_cpython_internal_calls(self, frame, expected):
        # GIVEN/WHEN/THEN
        assert is_cpython_internal(frame) is expected

    @pytest.mark.parametrize(
        "frame, expected",
        [
            [("somefunc", "runpy.py", 100), False],
            [
                (
                    "_PyEval_EvalFrameDefault",
                    "/src/python/python3.8/Python/ceval.c",
                    100,
                ),
                False,
            ],
            # Python 3.14 tail call interpreter frames should not be interesting
            [
                (
                    "_TAIL_CALL_INSTRUMENTED_CALL.llvm.10282351651392433962",
                    "/src/python/python3.14/Python/ceval.c",
                    100,
                ),
                False,
            ],
            [
                (
                    "PyArg_ParseTuple",
                    "/src/python/python3.8/Python/ceval.c",
                    100,
                ),
                True,
            ],
            [("somefunc", "myapp.py", 100), True],
            [
                (
                    "_PyEval_CompileCode",
                    "/src/python/python3.8/Include/code.h",
                    100,
                ),
                False,
            ],
        ],
    )
    def test_frame_interesting(self, frame, expected):
        # GIVEN/WHEN/THEN
        assert is_frame_interesting(frame) is expected

    @pytest.mark.parametrize(
        "frame, expected",
        [
            [("somefunc", "runpy.py", 100), False],
            [
                (
                    "_PyEval_EvalFrameDefault",
                    "/src/python/python3.8/Python/ceval.c",
                    100,
                ),
                False,
            ],
            [("somefunc", "<frozen importlib._blabla>", 100), True],
            [("somefunc", "<frozen something else>", 13), False],
            [("somefunc", "<frozen importlib>", 23), True],
            [("somefunc", "<frozen something._blich>", 11), False],
            [("somefunc", "ceval.c", 11), False],
            [("import_name", "ceval.c", 131), True],
            [("import_from", "ceval.c", 21), True],
            [("import_all_from", "ceval.c", 1), True],
            [("import_name", "otherfile.c", 14), False],
            [("import_from", "otherfile.c", 13), False],
            [("import_all_from", "otherfile.c", 12), False],
        ],
    )
    def test_is_frame_from_import_system(self, frame, expected):
        # GIVEN/WHEN/THEN
        assert is_frame_from_import_system(frame) is expected



import contextlib
import json
import os
import platform
import pty
import re
import signal
import subprocess
import sys
import sysconfig
import textwrap
import time
from pathlib import Path
from unittest.mock import patch

import pytest

from memray import FileFormat
from memray import Tracker
from memray._test import MemoryAllocator
from memray._test import set_thread_name
from memray.commands import main

TIMEOUT = 10


@pytest.fixture
def simple_test_file(tmp_path):
    code_file = tmp_path / "code.py"
    program = textwrap.dedent(
        """\
        from memray._test import MemoryAllocator
        print("Allocating some memory!")
        allocator = MemoryAllocator()
        allocator.valloc(1024)
        """
    )
    code_file.write_text(program)
    yield code_file


@pytest.fixture
def test_file_returns_from_fork(tmp_path):
    code_file = tmp_path / "code.py"
    program = textwrap.dedent(
        """\
        import os
        os.fork()
        """
    )
    code_file.write_text(program)
    yield code_file


@contextlib.contextmanager
def track_and_wait(output_dir, sleep_after=100):
    """Creates a test script which does some allocations, and upon leaving the context manager,
    it blocks until the allocations have completed."""

    fifo = output_dir / "snapshot_taken.event"
    os.mkfifo(fifo)

    program = textwrap.dedent(
        f"""\
        import time
        from memray._test import MemoryAllocator
        allocator = MemoryAllocator()
        allocator.valloc(1024)
        allocator.free()
        with open("{fifo}", "w") as fifo:
            fifo.write("done")
        time.sleep({sleep_after})
        """
    )
    program_file = output_dir / "file.py"
    program_file.write_text(program)
    yield program_file

    # Wait until we are tracking
    with open(fifo, "r") as f:
        assert f.read() == "done"


def _wait_until_process_blocks(pid: int) -> None:
    # Sleep longer on GitHub Actions macOS runners to reduce test flakiness
    default_sleep = 5.0 if "GITHUB_ACTIONS" in os.environ else 1.0
    if "linux" not in sys.platform:
        time.sleep(default_sleep)
        return
    # Signal numbers from https://filippo.io/linux-syscall-table/
    arch = platform.machine()
    if arch == "x86_64":
        sleep_syscall = "35"
        connect_syscall = "42"
        accept_syscall = "43"
        clock_nanosleep = "230"
    elif arch == "aarch64":
        sleep_syscall = "101"
        connect_syscall = "203"
        accept_syscall = "202"
        clock_nanosleep = "115"
    else:
        # No idea what syscalls numbers to wait on, so we will just
        # sleep for a long enough period and hope for the best
        time.sleep(default_sleep)
        return
    syscalls_to_wait = {sleep_syscall, clock_nanosleep, connect_syscall, accept_syscall}
    current_syscall = ""
    while True:
        syscall = Path(f"/proc/{pid}/syscall")
        current_syscall, *_ = syscall.read_text().split()
        if current_syscall.strip() in syscalls_to_wait:
            return
        time.sleep(0.1)


def generate_sample_results(
    tmp_path,
    code,
    *,
    native=False,
    trace_python_allocators=False,
    disable_pymalloc=False,
):
    results_file = tmp_path / "result.bin"
    env = os.environ.copy()
    env["PYTHONMALLOC"] = "malloc" if disable_pymalloc else "default"
    subprocess.run(
        [
            sys.executable,
            "-m",
            "memray",
            "run",
            *(["--native"] if native else []),
            *(["--trace-python-allocators"] if trace_python_allocators else []),
            "--output",
            str(results_file),
            str(code),
        ],
        cwd=str(tmp_path),
        check=True,
        capture_output=True,
        text=True,
        env=env,
    )
    return results_file, code


class TestRunSubcommand:
    def test_run(self, tmp_path, simple_test_file):
        # GIVEN / WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                str(simple_test_file),
            ],
            check=True,
            capture_output=True,
            text=True,
            cwd=str(tmp_path),
        )

        # THEN
        assert "Allocating some memory!" in proc.stdout
        assert proc.returncode == 0
        assert "example commands" in proc.stdout

        out_file = re.search("Writing profile results into (.*)", proc.stdout).group(1)
        assert (tmp_path / out_file).exists()

    def test_run_override_output(self, tmp_path, simple_test_file):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--output",
                str(out_file),
                str(simple_test_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert "Allocating some memory!" in proc.stdout
        assert proc.returncode == 0
        assert out_file.exists()

    def test_run_overwrite_output_file(self, tmp_path, simple_test_file):
        # GIVEN
        out_file = tmp_path / "result.bin"
        out_file.write_bytes(b"oops" * 1024 * 1024)
        assert out_file.stat().st_size == 4 * 1024 * 1024
        assert out_file.read_bytes()[:4] == b"oops"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--force",
                "--output",
                str(out_file),
                str(simple_test_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert "Allocating some memory!" in proc.stdout
        assert proc.returncode == 0
        assert 0 < out_file.stat().st_size < 4 * 1024 * 1024
        assert out_file.read_bytes()[:4] != b"oops"

    @pytest.mark.parametrize("compress", [True, False])
    def test_run_buffered_file_io(self, tmp_path, simple_test_file, compress):
        # GIVEN
        from memray import FileReader

        out_file = tmp_path / "result.bin"
        args = [
            sys.executable,
            "-m",
            "memray",
            "run",
            "--buffered-file-io",
            "--output",
            str(out_file),
        ]
        if not compress:
            args.append("--no-compress")
        args.append(str(simple_test_file))

        # WHEN
        proc = subprocess.run(
            args,
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert "Allocating some memory!" in proc.stdout
        assert proc.returncode == 0
        assert out_file.exists()

        # The capture file written through the buffered sink (including the
        # header rewritten via seek at shutdown) must be readable.
        records = list(FileReader(str(out_file)).get_allocation_records())
        assert any(record.size == 1024 for record in records)

    def test_run_file_with_args(self, tmp_path):
        """Execute a Python script and make sure the arguments in the script
        are correctly forwarded."""

        # GIVEN
        out_file = tmp_path / "result.bin"
        target_file = tmp_path / "test.py"
        target_file.write_text(
            textwrap.dedent(
                """\
        import sys
        print(f"Command: {sys.argv[0]}")
        print(f"Arg: {sys.argv[1]}")
        """
            )
        )

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--output",
                str(out_file),
                str(target_file),
                "arg1",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 0
        assert re.search(r"Command: (.*)test\.py", proc.stdout)
        assert "Arg: arg1" in proc.stdout
        assert out_file.exists()

    def test_sys_manipulations_when_running_script(self, tmp_path):
        # GIVEN
        out_file = tmp_path / "result.bin"
        target_file = tmp_path / "test.py"
        target_file.write_text("import json, sys; print(json.dumps(sys.path))")

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                str(target_file),
                "some",
                "provided args",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -m` put cwd in sys.path; ensure we replaced it.
        path = json.loads(proc.stdout)
        assert os.getcwd() not in path
        assert str(tmp_path) in path

    @pytest.mark.parametrize(
        "isolation_flag", ["-I"] + (["-P"] if sys.version_info > (3, 11) else [])
    )
    def test_suppressing_sys_manipulations_when_running_script(
        self, tmp_path, isolation_flag
    ):
        # GIVEN
        out_file = tmp_path / "result.bin"
        target_file = tmp_path / "test.py"
        target_file.write_text("import json, sys; print(json.dumps(sys.path))")

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                isolation_flag,
                "-m",
                "memray",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                str(target_file),
                "some",
                "provided args",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -m` did not put cwd in sys.path; ensure it isn't
        # there, and neither is the tmp_path we would have replaced it with.
        path = json.loads(proc.stdout)
        assert os.getcwd() not in path
        assert str(tmp_path) not in path

    def test_sys_manipulations_when_running_module(self, tmp_path):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-c",
                "import sys; from memray.commands import main; sys.exit(main())",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                "-m",
                "site",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -c` put "" in sys.path; ensure we replaced it.
        path = eval(
            " ".join(line for line in proc.stdout.splitlines() if line.startswith(" "))
        )
        assert "" not in path
        assert os.getcwd() in path

    @pytest.mark.parametrize(
        "isolation_flag", ["-I"] + (["-P"] if sys.version_info > (3, 11) else [])
    )
    def test_suppressing_sys_manipulations_when_running_module(
        self, tmp_path, isolation_flag
    ):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                isolation_flag,
                "-c",
                "import sys; from memray.commands import main; sys.exit(main())",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                "-m",
                "site",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -c` did not put "" in sys.path; ensure it isn't
        # there, and neither is the os.getcwd() we would have replaced it with.
        path = eval(
            " ".join(line for line in proc.stdout.splitlines() if line.startswith(" "))
        )
        assert "" not in path
        assert os.getcwd() not in path

    def test_sys_manipulations_when_running_cmd(self, tmp_path):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                "-c",
                "import json, sys; print(json.dumps(sys.path))",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -m` put cwd in sys.path; ensure we replaced it.
        path = json.loads(proc.stdout)
        assert os.getcwd() not in path
        assert "" in path

    @pytest.mark.parametrize(
        "isolation_flag", ["-I"] + (["-P"] if sys.version_info > (3, 11) else [])
    )
    def test_suppressing_sys_manipulations_when_running_cmd(
        self, tmp_path, isolation_flag
    ):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                isolation_flag,
                "-m",
                "memray",
                "run",
                "--quiet",
                "--output",
                str(out_file),
                "-c",
                "import json, sys; print(json.dumps(sys.path))",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert out_file.exists()
        assert proc.returncode == 0
        # Running `python -m` did not put cwd in sys.path; ensure it isn't
        # there, and neither is the "" we would have replaced it with.
        path = json.loads(proc.stdout)
        assert os.getcwd() not in path
        assert "" not in path

    @pytest.mark.parametrize("option", [None, "--live", "--live-remote"])
    def test_run_file_that_is_not_python(self, capsys, option):
        """Execute a non-Python script and make sure that we raise a good error"""

        # GIVEN / WHEN
        assert main(["run", *([option] if option else ()), sys.executable]) == 1

        # THEN
        captured = capsys.readouterr()
        assert (
            captured.err.strip()
            == "Only valid Python files or commands can be executed under memray"
        )

    @patch("memray.commands.run.os.getpid")
    @pytest.mark.parametrize("flag", [None, "--buffered-file-io"])
    def test_run_file_exists(self, getpid, tmp_path, monkeypatch, capsys, flag):
        # GIVEN / WHEN
        getpid.return_value = 0
        (tmp_path / "memray-json.tool.0.bin").touch()
        monkeypatch.chdir(tmp_path)

        # THEN
        assert main(["run", *([flag] if flag else ()), "-m", "json.tool", "-h"]) == 1
        captured = capsys.readouterr()
        expected_error = (
            "Output file memray-json.tool.0.bin already exists. Memray can overwrite it"
            " with the --force CLI argument or the overwrite=True API argument."
        )
        assert captured.err.strip() == expected_error

    def test_run_output_file_directory_does_not_exist(self, capsys):
        # GIVEN / WHEN / THEN

        assert main(["run", "--output", "/doesn/t/exist", "-m", "json.tool", "-h"]) == 1
        captured = capsys.readouterr()
        assert (
            captured.err.strip()
            == "Could not create output file /doesn/t/exist: No such file or directory"
        )

    @pytest.mark.parametrize("quiet", [True, False])
    def test_quiet(self, quiet, tmp_path, simple_test_file):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                *(["-q"] if quiet else []),
                "--output",
                str(out_file),
                str(simple_test_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 0
        assert out_file.exists()
        if quiet:
            assert str(out_file) not in proc.stdout
        else:
            assert str(out_file) in proc.stdout

    def test_not_quiet_and_fork(self, tmp_path, test_file_returns_from_fork):
        # GIVEN
        out_file = tmp_path / "result.bin"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--output",
                str(out_file),
                str(test_file_returns_from_fork),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 0
        assert out_file.exists()
        assert proc.stdout.count("Some example commands to generate reports") == 1


class TestParseSubcommand:
    def test_successful_parse(self, tmp_path):
        # GIVEN
        record_types = [
            "ALLOCATION",
            "MEMORY_MAP_START",
            "SEGMENT_HEADER",
            "SEGMENT",
            "NATIVE_FRAME_ID",
            "FRAME_PUSH",
            "FRAME_POP",
            "MEMORY_RECORD",
            "CONTEXT_SWITCH",
            "CODE_OBJECT",
            "TRAILER",
        ]

        code_file = tmp_path / "code.py"
        program = textwrap.dedent(
            """\
            import time
            from memray._test import MemoryAllocator
            print("Allocating some memory!")
            allocator = MemoryAllocator()

            def foo():
                allocator.valloc(1024)
                allocator.free()

            def bar():
                allocator.valloc(2048)
                allocator.free()

            # Ensure we see multiple allocation locations to guarantee
            # that there's at least one FRAME_POP record.
            foo()
            bar()

            # Give it time to generate some memory records
            time.sleep(0.1)
            """
        )
        code_file.write_text(program)
        record_count_by_type = dict.fromkeys(record_types, 0)
        results_file, _ = generate_sample_results(tmp_path, code_file, native=True)

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "parse",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
            cwd=str(tmp_path),
        )

        # THEN
        _, *records = proc.stdout.splitlines()

        for record in records:
            record_count_by_type[record.partition(" ")[0]] += 1

        for record_type, count in record_count_by_type.items():
            assert (
                count > 0
            ), f"no {record_type} records found in {record_count_by_type}"

    def test_successful_parse_of_aggregated_capture_file(self, tmp_path):
        # GIVEN
        results_file = tmp_path / "result.bin"
        record_types = [
            "MEMORY_SNAPSHOT",
            "AGGREGATED_ALLOCATION",
            "PYTHON_TRACE_INDEX",
            "PYTHON_FRAME_INDEX",
            "NATIVE_TRACE_INDEX",
            "MEMORY_MAP_START",
            "SEGMENT_HEADER",
            "SEGMENT",
            "AGGREGATED_TRAILER",
            "CODE_OBJECT",
        ]

        with Tracker(
            results_file,
            native_traces=True,
            file_format=FileFormat.AGGREGATED_ALLOCATIONS,
        ):
            if set_thread_name("main") == 0:
                # We should get CONTEXT_SWITCH and THREAD_RECORD records only
                # if we can set the thread name. On macOS, we won't get these.
                record_types.append("CONTEXT_SWITCH")
                record_types.append("THREAD_RECORD")

            allocator = MemoryAllocator()
            allocator.valloc(1024)
            allocator.free()
            # Give it time to generate some memory records
            time.sleep(0.1)

        record_count_by_type = dict.fromkeys(record_types, 0)

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "parse",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
            cwd=str(tmp_path),
        )

        # THEN
        _, *records = proc.stdout.splitlines()
        print(records)

        for record in records:
            record_count_by_type[record.partition(" ")[0]] += 1

        for record_type, count in record_count_by_type.items():
            assert (
                count > 0
            ), f"no {record_type} records found in {record_count_by_type}"

    def test_error_when_stdout_is_a_tty(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        _, controlled = pty.openpty()

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "parse",
                str(results_file),
            ],
            stdout=controlled,
            stderr=subprocess.PIPE,
            text=True,
            cwd=str(tmp_path),
        )

        # THEN
        assert "You must redirect stdout" in proc.stderr
        assert proc.returncode == 1

    def test_error_when_input_file_does_not_exist(self, tmp_path):
        # GIVEN
        results_file = tmp_path / "does/not/exist"

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "parse",
                results_file,
            ],
            capture_output=True,
            text=True,
            cwd=str(tmp_path),
        )

        # THEN
        assert "Reason: No such file" in proc.stderr
        assert proc.returncode == 1


class TestFlamegraphSubCommand:
    def test_reads_from_correct_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                str(results_file),
            ],
            cwd=str(tmp_path),
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-flamegraph-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    def test_no_web_embeds_d3_assets(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                "--no-web",
                str(results_file),
            ],
            cwd=str(tmp_path),
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-flamegraph-result.html"
        html = output_file.read_text()
        assert '<script src="https://' not in html
        assert '<link rel="stylesheet" href="https://' not in html
        assert "createPopper" not in html
        assert "d3.schemePastel1" in html

    def test_can_generate_reports_with_native_traces(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                str(results_file),
            ],
            cwd=str(tmp_path),
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-flamegraph-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    def test_writes_to_correct_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "output.html"

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                str(results_file),
                "--output",
                str(output_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    def test_output_file_already_exists(self, tmp_path, simple_test_file, monkeypatch):
        """Check that when the output file is derived form the input name, we
        fail when there is already a file with the same name as the output."""

        # GIVEN
        monkeypatch.chdir(tmp_path)
        # This will generate "result.bin"
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "memray-flamegraph-result.html"
        output_file.touch()

        # WHEN
        ret = main(["flamegraph", str(results_file)])

        # THEN
        assert ret != 0

    def test_split_threads_subcommand(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "output.html"

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                "--split-threads",
                str(results_file),
                "--output",
                str(output_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    @pytest.mark.parametrize("trace_python_allocators", [True, False])
    @pytest.mark.parametrize("disable_pymalloc", [True, False])
    def test_leaks_with_pymalloc_warning(
        self,
        tmp_path,
        simple_test_file,
        trace_python_allocators,
        disable_pymalloc,
    ):
        if disable_pymalloc and sysconfig.get_config_var("Py_GIL_DISABLED"):
            pytest.skip("PYTHONMALLOC=malloc is not supported in free-threading builds")

        # GIVEN
        results_file, _ = generate_sample_results(
            tmp_path,
            simple_test_file,
            native=True,
            trace_python_allocators=trace_python_allocators,
            disable_pymalloc=disable_pymalloc,
        )
        output_file = tmp_path / "output.html"
        warning_expected = not trace_python_allocators and not disable_pymalloc

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "flamegraph",
                "--leaks",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-flamegraph-result.html"
        assert output_file.exists()
        assert warning_expected == (
            "Report generated using <code>--leaks</code> with an arena allocator"
            in output_file.read_text()
        )


class TestSummarySubCommand:
    def test_summary_generated(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "summary",
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert output

    def test_temporary_allocations_summary(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)

        # WHEN
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "summary",
                "--temporary-allocations",
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert output


class TestTreeSubCommand:
    def test_tree_generated(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        env = os.environ.copy()
        env["TEXTUAL_PRESS"] = "q"

        # WHEN
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "tree",
                str(results_file),
            ],
            cwd=str(tmp_path),
            stderr=subprocess.DEVNULL,
            text=True,
            env=env,
        )

        # THEN
        assert output

    def test_temporary_allocations_tree(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        env = os.environ.copy()
        env["TEXTUAL_PRESS"] = "q"

        # WHEN
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "tree",
                "--temporary-allocations",
                str(results_file),
            ],
            cwd=str(tmp_path),
            stderr=subprocess.DEVNULL,
            text=True,
            env=env,
        )

        # THEN
        assert output


class TestStatsSubCommand:
    def test_report_generated(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        output = subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert "VALLOC" in output

    def test_json_generated(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        json_file = tmp_path / "memray-stats-result.bin.json"

        # WHEN
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                "--json",
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert json_file.exists()
        assert isinstance(json.loads(json_file.read_text()), dict)

    def test_json_generated_to_pretty_file_name(self, tmp_path, simple_test_file):
        # GIVEN
        orig_results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        results_file = orig_results_file.with_name("memray-foobar.bin")
        orig_results_file.rename(results_file)
        json_file = tmp_path / "memray-stats-foobar.bin.json"

        # WHEN
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                "--json",
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert json_file.exists()
        assert isinstance(json.loads(json_file.read_text()), dict)

    def test_json_generated_to_known_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        json_file = tmp_path / "output.json"

        # WHEN
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                "--json",
                "-o",
                str(json_file),
                str(results_file),
            ],
            cwd=str(tmp_path),
            text=True,
        )

        # THEN
        assert json_file.exists()
        assert isinstance(json.loads(json_file.read_text()), dict)

    def test_json_generated_to_existing_known_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        json_file = tmp_path / "output.json"
        json_file.write_text("oops")

        # WHEN
        try:
            exc = None
            subprocess.check_output(
                [
                    sys.executable,
                    "-m",
                    "memray",
                    "stats",
                    "--json",
                    "-o",
                    str(json_file),
                    str(results_file),
                ],
                cwd=str(tmp_path),
                stderr=subprocess.PIPE,
                text=True,
            )
        except subprocess.CalledProcessError as e:
            exc = e

        # THEN
        assert exc is not None
        assert "File already exists, will not overwrite" in exc.stderr

    def test_json_overwrites_existing_known_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, _ = generate_sample_results(tmp_path, simple_test_file)
        json_file = tmp_path / "output.json"
        json_file.write_text("oops")

        # WHEN
        subprocess.check_output(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                "--json",
                "--force",
                "--output",
                str(json_file),
                str(results_file),
            ],
            cwd=str(tmp_path),
            stderr=subprocess.PIPE,
            text=True,
        )

        # THEN
        assert json_file.exists()
        assert isinstance(json.loads(json_file.read_text()), dict)

    def test_report_detects_corrupt_input(self, tmp_path):
        # GIVEN
        bad_file = Path(tmp_path) / "badfile.bin"
        bad_file.write_text("This is some garbage")

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "stats",
                str(bad_file),
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 1
        assert re.match(r"Failed to compute statistics for .*badfile\.bin", proc.stderr)


class TestTableSubCommand:
    def test_reads_from_correct_file(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "table",
                str(results_file),
            ],
            cwd=str(tmp_path),
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-table-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    def test_split_threads(self, tmp_path, simple_test_file):
        # GIVEN
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "table",
                "--split-threads",
                str(results_file),
            ],
            cwd=str(tmp_path),
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-table-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()


class TestReporterSubCommands:
    @pytest.mark.parametrize(
        "report", ["flamegraph", "table", "summary", "tree", "stats"]
    )
    def test_report_detects_missing_input(self, report):
        # GIVEN / WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                "nosuchfile",
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 1
        assert "No such file: nosuchfile" in proc.stderr

    @pytest.mark.parametrize("report", ["flamegraph", "table", "summary", "tree"])
    def test_report_detects_corrupt_input(self, tmp_path, report):
        # GIVEN
        bad_file = Path(tmp_path) / "badfile.bin"
        bad_file.write_text("This is some garbage")

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                str(bad_file),
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 1
        assert re.match(
            r"Failed to parse allocation records in .*badfile\.bin", proc.stderr
        )

    @pytest.mark.parametrize("report", ["flamegraph", "table"])
    def test_report_leaks_argument(self, tmp_path, simple_test_file, report):
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "output.html"

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                "--leaks",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / f"memray-{report}-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    @pytest.mark.parametrize("report", ["flamegraph", "table"])
    def test_report_temporary_allocations_argument(
        self, tmp_path, simple_test_file, report
    ):
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "output.html"

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                str(results_file),
                "--temporary-allocations",
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / f"memray-{report}-result.html"
        assert output_file.exists()
        assert str(source_file) in output_file.read_text()

    @pytest.mark.parametrize("report", ["flamegraph", "table"])
    def test_report_incompatible_arguments(self, tmp_path, simple_test_file, report):
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                "--temporary-allocations",
                "--leaks",
                str(results_file),
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode != 0
        assert (
            "--leaks: not allowed with argument --temporary-allocations" in proc.stderr
        )

    @pytest.mark.parametrize("report", ["flamegraph", "table", "summary", "tree"])
    def test_report_both_temporary_allocation_arguments(
        self, tmp_path, simple_test_file, report
    ):
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                report,
                "--temporary-allocations",
                "--temporary-allocation-threshold=1",
                str(results_file),
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode != 0
        assert (
            "--temporary-allocation-threshold: not allowed with"
            " argument --temporary-allocations" in proc.stderr
        )


class TestLiveRemoteSubcommand:
    def test_live_tracking(self, tmp_path, simple_test_file, free_port):
        # GIVEN
        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--live-remote",
                "--live-port",
                str(free_port),
                str(simple_test_file),
            ],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
        )

        _wait_until_process_blocks(server.pid)

        client = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "live",
                str(free_port),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
            stdin=subprocess.PIPE,
        )

        # WHEN

        try:
            server.communicate(timeout=TIMEOUT)
            client.communicate(b"q", timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            server.terminate()
            client.terminate()
            server.wait(timeout=TIMEOUT)
            client.wait(timeout=TIMEOUT)
            raise

        server.communicate()

        # THEN
        assert server.returncode == 0
        assert client.returncode == 0

    def test_live_tracking_waits_for_client(self, simple_test_file):
        # GIVEN
        env = os.environ.copy()
        env["PYTHONUNBUFFERED"] = "1"

        # WHEN
        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--live-remote",
                str(simple_test_file),
            ],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # THEN
        assert b"another shell to see live results\n" in server.stdout.readline()
        server.terminate()
        server.wait(timeout=TIMEOUT)

    @pytest.mark.parametrize("port", [0, 2**16, 1000000])
    def test_run_live_tracking_invalid_port(self, simple_test_file, port):
        # GIVEN
        env = os.environ.copy()
        env["PYTHONUNBUFFERED"] = "1"

        # WHEN
        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--live-remote",
                "--live-port",
                str(port),
                str(simple_test_file),
            ],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        # THEN
        assert "Invalid port" in server.stderr.readline()
        server.terminate()
        server.wait(timeout=TIMEOUT)

    @pytest.mark.parametrize("port", [0, 2**16, 1000000])
    def test_live_tracking_invalid_port(self, port):
        # GIVEN
        env = os.environ.copy()
        env["PYTHONUNBUFFERED"] = "1"

        # WHEN
        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "live",
                str(port),
            ],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        # THEN
        assert "Invalid port" in server.stderr.readline()
        server.terminate()
        server.wait(timeout=TIMEOUT)

    def test_live_tracking_server_when_client_disconnects(self, free_port, tmp_path):
        # GIVEN
        env = os.environ.copy()
        env["PYTHONUNBUFFERED"] = "1"

        test_file = tmp_path / "test.py"
        test_file.write_text("import time; time.sleep(3)")

        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--live-port",
                str(free_port),
                "--live-remote",
                str(test_file),
            ],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )

        _wait_until_process_blocks(server.pid)

        client = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "live",
                str(free_port),
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.DEVNULL,
        )

        # WHEN
        try:
            client.communicate(b"q", timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            client.terminate()
            client.wait(timeout=TIMEOUT)
            raise

        try:
            _, stderr = server.communicate(timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            server.terminate()
            server.wait(timeout=TIMEOUT)
            raise

        # THEN
        assert "Encountered error in 'send' call:" not in stderr
        assert "Traceback" not in stderr

    def test_live_tracking_server_exits_properly_on_sigint(self, simple_test_file):
        # GIVEN
        env = os.environ.copy()
        env["PYTHONUNBUFFERED"] = "1"

        server = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "run",
                "--live-remote",
                str(simple_test_file),
            ],
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            # Explicitly reset the signal handler for SIGINT to work around any signal
            # masking that might happen on Jenkins.
            preexec_fn=lambda: signal.signal(signal.SIGINT, signal.default_int_handler),
        )

        # WHEN
        server.stdout.readline()  # wait for the startup message
        # Ensure that it's waiting on the socket
        _wait_until_process_blocks(server.pid)

        server.send_signal(signal.SIGINT)
        try:
            _, stderr = server.communicate(timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            server.kill()
            server.wait(timeout=TIMEOUT)
            raise

        # THEN
        assert server.returncode == 0
        assert b"Exception ignored" not in stderr
        assert b"Traceback (most recent call last):" not in stderr
        assert b"Interrupted system call" not in stderr

    def test_live_client_exits_properly_on_sigint_before_connecting(self, free_port):
        # GIVEN
        client = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "memray",
                "live",
                str(free_port),
            ],
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            # Explicitly reset the signal handler for SIGINT to work around any signal
            # masking that might happen on Jenkins.
            preexec_fn=lambda: signal.signal(signal.SIGINT, signal.default_int_handler),
        )

        # Ensure that it's waiting on the socket
        _wait_until_process_blocks(client.pid)

        # WHEN
        client.send_signal(signal.SIGINT)
        try:
            client.wait(timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            client.terminate()

        # THEN
        assert client.returncode == 0


class TestLiveSubcommand:
    def test_live_tracking(self, tmp_path):
        # GIVEN
        with track_and_wait(tmp_path) as program_file:
            server = subprocess.Popen(
                [
                    sys.executable,
                    "-m",
                    "memray",
                    "run",
                    "--live",
                    str(program_file),
                ],
                stdin=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )

        # WHEN
        try:
            server.communicate(b"q", timeout=TIMEOUT)
        except subprocess.TimeoutExpired:
            server.kill()
            raise

        # THEN
        assert server.returncode == 0


class TestTransformSubCommands:
    def test_report_detects_missing_input(self):
        # GIVEN / WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "transform",
                "gprof2dot",
                "nosuchfile",
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 1
        assert "No such file: nosuchfile" in proc.stderr

    def test_report_detects_corrupt_input(self, tmp_path):
        # GIVEN
        bad_file = Path(tmp_path) / "badfile.bin"
        bad_file.write_text("This is some garbage")

        # WHEN
        proc = subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "transform",
                "gprof2dot",
                str(bad_file),
            ],
            capture_output=True,
            text=True,
        )

        # THEN
        assert proc.returncode == 1
        assert re.match(
            r"Failed to parse allocation records in .*badfile\.bin", proc.stderr
        )

    def test_report_leaks_argument(self, tmp_path, simple_test_file):
        results_file, source_file = generate_sample_results(
            tmp_path, simple_test_file, native=True
        )
        output_file = tmp_path / "output.html"

        # WHEN
        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "transform",
                "gprof2dot",
                "--leaks",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        # THEN
        output_file = tmp_path / "memray-gprof2dot-result.json"
        assert output_file.exists()
        output_text = output_file.read_text()
        if "<unknown stack>" in output_text:
            pytest.xfail("Hybrid stack generation is not fully working")
        assert str(source_file) in output_text

    def test_report_speedscope_argument(self, tmp_path, simple_test_file):
        results_file, _ = generate_sample_results(
            tmp_path, simple_test_file, native=False
        )

        subprocess.run(
            [
                sys.executable,
                "-m",
                "memray",
                "transform",
                "speedscope",
                str(results_file),
            ],
            check=True,
            capture_output=True,
            text=True,
        )

        output_file = tmp_path / "memray-speedscope-result.speedscope.json"
        assert output_file.exists()

        output_text = output_file.read_text()
        output_data = json.loads(output_text)
        assert output_data["$schema"] == (
            "https://www.speedscope.app/file-format-schema.json"
        )
        assert [profile["type"] for profile in output_data["profiles"]] == [
            "sampled",
            "sampled",
        ]
        assert output_data["shared"]["frames"]



import subprocess

from setuptools import Extension
from setuptools import setup

# Compile the shared library before building the extension
subprocess.run(
    ["gcc", "-shared", "-o", "sharedlibs/sharedlib.so", "sharedlibs/sharedlib.c"]
)


setup(
    name="ext",
    version="1.0",
    ext_modules=[
        Extension(
            "ext",
            sources=["ext.c"],
            extra_link_args=["-Wl,-rpath,$ORIGIN/sharedlibs"],
        )
    ],
)



import numpy as np

# DO NOT CHANGE
MB_CONVERSION = 1024 * 1024
DUPLICATE_CONST = 2

SIZE_OF_DATA_IN_MB = 100
SUBTRACT_AMOUNT = 3
POWER_AMOUNT = 2
ADD_AMOUNT = 10
# DO NOT CHANGE


def load_xMb_of_data(mb: int) -> np.ndarray:
    size = MB_CONVERSION * mb
    return np.ones(size, dtype=np.uint8)


def duplicate_data(data: np.ndarray) -> np.ndarray:
    return data * DUPLICATE_CONST


def add_scalar(data: np.ndarray, scalar: int) -> np.ndarray:
    return data + scalar


def subtract_scalar(data: np.ndarray, scalar: int) -> np.ndarray:
    return data - scalar


def raise_to_power(data: np.ndarray, power: int) -> np.ndarray:
    return np.power(data, power)


def process_data_fix_1():
    # no extra reference to the original array
    return add_scalar(
        duplicate_data(
            raise_to_power(
                subtract_scalar(load_xMb_of_data(SIZE_OF_DATA_IN_MB), SUBTRACT_AMOUNT),
                POWER_AMOUNT,
            )
        ),
        ADD_AMOUNT,
    )


def process_data_fix_2():
    # reusing the local variable instead of allocating more space
    # this approach is called 'hidden mutability'
    data = load_xMb_of_data(SIZE_OF_DATA_IN_MB)
    data = subtract_scalar(data, SUBTRACT_AMOUNT)
    data = raise_to_power(data, POWER_AMOUNT)
    data = duplicate_data(data)
    data = add_scalar(data, ADD_AMOUNT)
    return data


# Use these to select which solution to test
process_data = process_data_fix_1

if __name__ == "__main__":
    process_data()



import os

import pytest
from IPython.core.displaypub import CapturingDisplayPublisher
from IPython.core.interactiveshell import InteractiveShell
from traitlets.config import Config


def run_in_ipython_shell(tmpdir, cells):
    """Run the given cells in an IPython shell and return the HTML output."""
    InteractiveShell.clear_instance()

    config = Config()
    config.HistoryAccessor.enabled = False  # Disable sqlite history
    shell = InteractiveShell.instance(
        display_pub_class=CapturingDisplayPublisher,
        config=config,
    )
    prev_running_dir = os.getcwd()
    try:
        os.chdir(tmpdir)
        for cell in cells:
            shell.run_cell(cell)
    finally:
        os.chdir(prev_running_dir)

    InteractiveShell.clear_instance()
    try:
        html = shell.display_pub.outputs[-1]["data"]["text/html"]
        return html
    except IndexError:
        return None


@pytest.mark.filterwarnings("ignore")
class TestIPython:
    def test_ipython_profiling(self, tmpdir):
        """Test that the IPython extension works."""
        # GIVEN
        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph
            x = "a" * 10000
            """,
        ]

        # WHEN

        html = run_in_ipython_shell(tmpdir, code)

        # THEN

        assert html is not None
        assert "<iframe" in html
        assert "flamegraph.html" in html

    def test_exception_while_ipython_profiling(self, tmpdir):
        """Test that the IPython extension works even if an exception
        is raised."""
        # GIVEN

        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph
            x = "a" * 10000
            1/0
            """,
        ]

        # WHEN

        html = run_in_ipython_shell(tmpdir, code)

        # THEN

        assert html is not None
        assert "<iframe" in html
        assert "flamegraph.html" in html

    def test_passing_help_argument(self, tmpdir, capsys):
        # GIVEN

        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph -h
            """,
        ]

        # WHEN

        html = run_in_ipython_shell(tmpdir, code)

        # THEN

        assert html is None
        stdout, stderr = capsys.readouterr()
        assert "show this help message" in stdout
        assert "" == stderr

    def test_passing_invalid_argument(self, tmpdir, capsys):
        # GIVEN

        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph --oopsie
            """,
        ]

        # WHEN

        html = run_in_ipython_shell(tmpdir, code)

        # THEN

        assert html is None
        stdout, stderr = capsys.readouterr()
        assert "" == stdout
        assert "usage:" in stderr

    def test_passing_valid_arguments(self, tmpdir, capsys):
        # GIVEN

        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph --leaks --max-memory-records 5
            x = "a" * 10000
            """,
        ]

        # WHEN

        html = run_in_ipython_shell(tmpdir, code)

        # THEN

        assert html is not None
        stdout, _ = capsys.readouterr()
        assert "<iframe" in html
        assert "flamegraph.html" in html
        assert "Results saved to" in stdout

    @pytest.mark.parametrize(
        "args,title",
        [
            ("--temporary-allocation-threshold=2", "flamegraph report"),
            ("--leaks", "flamegraph report (memory leaks)"),
            ("--temporal", "temporal flamegraph report"),
            ("--leaks --temporal", "temporal flamegraph report (memory leaks)"),
        ],
    )
    def test_report_title_by_report_type(self, tmpdir, capsys, args, title):
        # GIVEN
        code = [
            "%load_ext memray",
            f"""
            %%memray_flamegraph {args}
            x = "a" * 10000
            """,
        ]

        # WHEN
        html = run_in_ipython_shell(tmpdir, code)

        # THEN
        assert html is not None
        stdout, _ = capsys.readouterr()
        assert "<iframe" in html
        assert "flamegraph.html" in html
        assert "Results saved to" in stdout

        title_tag = f"<title>memray - {title}</title>"
        assert title_tag in next(tmpdir.visit("**/flamegraph.html")).read()

    def test_passing_temporal_and_temporary_allocations(self, tmpdir, capsys):
        # GIVEN
        code = [
            "%load_ext memray",
            """
            %%memray_flamegraph --temporal --temporary-allocation-threshold=2
            """,
        ]

        # WHEN
        html = run_in_ipython_shell(tmpdir, code)

        # THEN
        assert html is None
        stdout, stderr = capsys.readouterr()
        assert "" == stdout
        assert "Can't create a temporal flame graph of temporary allocations" in stderr



"""Tests for exercising the public API."""

import pytest

from memray import FileDestination
from memray import FileFormat
from memray import FileReader
from memray import SocketDestination
from memray import Tracker
from memray._test import MemoryAllocator
from tests.utils import filter_relevant_allocations


def test_file_reader_as_context_manager(tmp_path):
    # GIVEN
    allocator = MemoryAllocator()
    result_file = tmp_path / "test.bin"
    # WHEN
    with Tracker(result_file):
        allocator.valloc(1234)
        allocator.free()

    # THEN
    with FileReader(result_file) as reader:
        pass

    with pytest.raises(ValueError, match="Operation on a closed FileReader"):
        list(reader.get_high_watermark_allocation_records())


def test_file_destination(tmp_path):
    # GIVEN
    allocator = MemoryAllocator()
    result_file = tmp_path / "test.bin"
    # WHEN
    with Tracker(destination=FileDestination(result_file)):
        allocator.valloc(1234)
        allocator.free()

    # THEN
    with FileReader(result_file) as reader:
        all_allocations = reader.get_allocation_records()
        vallocs_and_their_frees = list(filter_relevant_allocations(all_allocations))
        assert len(vallocs_and_their_frees) == 2


def test_file_destination_str_path(tmp_path):
    # GIVEN
    allocator = MemoryAllocator()
    result_file = str(tmp_path / "test.bin")
    # WHEN
    with Tracker(destination=FileDestination(result_file)):
        allocator.valloc(1234)
        allocator.free()

    # THEN
    with FileReader(result_file) as reader:
        all_allocations = reader.get_allocation_records()
        vallocs_and_their_frees = list(filter_relevant_allocations(all_allocations))
        assert len(vallocs_and_their_frees) == 2


def test_combine_destination_args():
    """Combining `writer` and `file_name` arguments in the `Tracker` should
    raise an exception."""

    # GIVEN/WHEN/THEN
    with pytest.raises(
        TypeError,
        match="Exactly one of 'file_name' or 'destination' argument must be specified",
    ):
        with Tracker(destination=SocketDestination(server_port=1234), file_name="foo"):
            pass


def test_no_destination_arg():
    """Not passing either `writer` or `file_name` argument in the `Tracker` should
    raise an exception."""

    # GIVEN/WHEN/THEN
    with pytest.raises(
        TypeError,
        match="Exactly one of 'file_name' or 'destination' argument must be specified",
    ):
        with Tracker():  # pragma: no cover
            pass


def test_follow_fork_with_socket_destination():
    # GIVEN
    with pytest.raises(RuntimeError, match="follow_fork requires an output file"):
        with Tracker(
            destination=SocketDestination(server_port=1234), follow_fork=True
        ):  # pragma: no cover
            pass


def test_aggregated_capture_with_socket_destination():
    # GIVEN
    with pytest.raises(
        RuntimeError, match="AGGREGATED_ALLOCATIONS requires an output file"
    ):
        with Tracker(
            destination=SocketDestination(server_port=1234),
            file_format=FileFormat.AGGREGATED_ALLOCATIONS,
        ):  # pragma: no cover
            pass



from pathlib import Path

import pytest

from memray import Tracker
from memray._test import rss_from_proc_status
from tests.utils import skip_if_macos


def test_two_trackers_cannot_be_active_at_the_same_time(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    output2 = Path(tmpdir) / "test2.bin"

    # WHEN
    with Tracker(output):
        # THEN
        with pytest.raises(RuntimeError):
            with Tracker(output2):
                pass


def test_the_same_tracker_cannot_be_activated_twice(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"

    # WHEN
    tracker = Tracker(output)
    with tracker:
        # Remove the file so we are not stopped by the file existence check
        output.unlink()
        # THEN
        with pytest.raises(RuntimeError):
            with tracker:
                pass


def test_thread_class_swap_during_tracking_does_not_crash(tmpdir, monkeypatch):
    """Regression test for https://github.com/bloomberg/memray/issues/856.

    ``gevent.monkey.patch_all()`` replaces ``threading.Thread`` with its own
    class while tracking is active. ``Tracker.__exit__`` must clean up the
    instrumentation it installed on the original ``Thread`` class, not the
    one that ``threading.Thread`` happens to point at on exit.
    """
    output = Path(tmpdir) / "test.bin"
    with Tracker(output):

        class FakeThread:
            pass

        monkeypatch.setattr("threading.Thread", FakeThread)


@skip_if_macos
def test_rss_from_proc_status_includes_hugetlb_pages():
    assert (
        rss_from_proc_status(
            "\n".join(
                (
                    "Name:\tpython",
                    "VmRSS:\t  128 kB",
                    "HugetlbPages:\t64 kB",
                )
            )
            + "\n"
        )
        == (128 + 64) * 1024
    )



import multiprocessing
import sys
from multiprocessing import Pool
from pathlib import Path

import pytest

from memray import AllocatorType
from memray import FileReader
from memray import Tracker
from memray._test import MemoryAllocator
from memray._test import PymallocDomain
from memray._test import PymallocMemoryAllocator
from tests.utils import filter_relevant_allocations
from tests.utils import filter_relevant_pymalloc_allocations


@pytest.fixture(scope="module", autouse=True)
def set_multiprocessing_to_fork():
    current_method = multiprocessing.get_start_method()
    multiprocessing.set_start_method("fork", force=True)
    yield
    multiprocessing.set_start_method(current_method, force=True)


def multiproc_func(repetitions):  # pragma: no cover
    allocator = MemoryAllocator()
    for _ in range(repetitions):
        allocator.valloc(1234)
        allocator.free()


def pymalloc_multiproc_func():  # pragma: no cover
    allocator = PymallocMemoryAllocator(PymallocDomain.PYMALLOC_RAW)
    allocator.calloc(1234)
    allocator.free()


@pytest.mark.no_cover
def test_allocations_with_multiprocessing(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    allocator = MemoryAllocator()

    # WHEN
    with Tracker(output):
        with Pool(3) as p:
            p.map(multiproc_func, [1, 10, 100, 1000, 2000, 3000, 4000, 5000])

        allocator.valloc(1234)
        allocator.free()

    relevant_records = list(
        filter_relevant_allocations(FileReader(output).get_allocation_records())
    )
    assert len(relevant_records) == 2

    vallocs = [
        record
        for record in relevant_records
        if record.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 1
    (valloc,) = vallocs
    assert valloc.size == 1234

    frees = [
        record for record in relevant_records if record.allocator == AllocatorType.FREE
    ]
    assert len(frees) == 1

    # No files created by child processes
    child_files = Path(tmpdir).glob("test.bin.*")
    assert list(child_files) == []


@pytest.mark.no_cover
def test_allocations_with_multiprocessing_following_fork(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    allocator = MemoryAllocator()

    # WHEN
    with Tracker(output, follow_fork=True):
        with Pool(3) as p:
            p.map(multiproc_func, [1, 10, 100, 1000, 2000, 3000, 4000, 5000])

        allocator.valloc(1234)
        allocator.free()

    # THEN
    relevant_records = list(
        filter_relevant_allocations(FileReader(output).get_allocation_records())
    )
    assert len(relevant_records) == 2

    vallocs = [
        record
        for record in relevant_records
        if record.allocator == AllocatorType.VALLOC
    ]
    assert len(vallocs) == 1
    (valloc,) = vallocs
    assert valloc.size == 1234

    frees = [
        record for record in relevant_records if record.allocator == AllocatorType.FREE
    ]
    assert len(frees) == 1

    child_files = Path(tmpdir).glob("test.bin.*")
    child_records = []
    for child_file in child_files:
        child_records.extend(
            filter_relevant_allocations(FileReader(child_file).get_allocation_records())
        )

    child_vallocs = [
        record for record in child_records if record.allocator == AllocatorType.VALLOC
    ]

    child_frees = [
        record for record in child_records if record.allocator == AllocatorType.FREE
    ]

    num_expected = 5000 + 4000 + 3000 + 2000 + 1000 + 100 + 10 + 1
    assert len(child_vallocs) == num_expected

    # We may be missing up to 1 free per process in the pool due to our
    # heuristic for recovering from truncated files (see the comment
    # where d_readable_size is set in source.cpp)
    assert 0 <= len(child_vallocs) - len(child_frees) <= 3
    for valloc in child_vallocs:
        assert valloc.size == 1234


@pytest.mark.no_cover
def test_pymalloc_allocations_after_fork(tmpdir):
    # GIVEN
    output = Path(tmpdir) / "test.bin"

    # WHEN
    with Tracker(output, follow_fork=True, trace_python_allocators=True):
        with Pool(3) as p:
            p.starmap(pymalloc_multiproc_func, [()] * 10)

    # THEN
    child_files = Path(tmpdir).glob("test.bin.*")
    child_records = []
    for child_file in child_files:
        child_records.extend(
            filter_relevant_pymalloc_allocations(
                FileReader(child_file).get_allocation_records(), size=1234
            )
        )

    print(child_records)
    child_callocs = [
        record
        for record in child_records
        if record.allocator == AllocatorType.PYMALLOC_CALLOC and record.size == 1234
    ]

    num_expected = 10
    assert len(child_callocs) == num_expected


@pytest.mark.no_cover
def test_stack_cleanup_after_fork(tmpdir):
    """Test that we don't crash miserably when we try to write pending Python
    frames when the profile function is deactivated if the tracker has been
    destroyed after a fork without `follow_fork=True`"""

    # GIVEN
    output = Path(tmpdir) / "test.bin"

    def foo():
        with Pool() as pool:
            result = pool.map_async(sys.setprofile, [None])
            return result.get(timeout=1)

    # WHEN

    with Tracker(output, follow_fork=False):
        result = foo()

    # THEN

    assert result == [None]



"""Tests to exercise socket-based read and write operations in the Tracker."""

import os
import subprocess
import sys
import textwrap
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator
from typing import Set
from typing import Tuple

import pytest

from memray import AllocatorType
from memray import SocketReader
from tests.utils import filter_relevant_allocations

TIMEOUT = 5
ALLOCATION_SIZE = 1234
MULTI_ALLOCATION_COUNT = 10

#
# Test helpers
#
_SCRIPT_TEMPLATE = """
import sys
from memray._test import MemoryAllocator
from memray._memray import SocketDestination
from memray._memray import Tracker

# Sanity checks
port = int(sys.argv[1])
assert sys.argv[2].endswith("allocations_made.event")
assert sys.argv[3].endswith("snapshot_taken.event")


def get_tracker():
    return Tracker(destination=SocketDestination(server_port=port))


def snapshot_point():
    print("[child] Notifying allocations made")
    with open(sys.argv[2], "w") as allocations_made:
        allocations_made.write("done")
    print("[child] Waiting on snapshot taken")
    with open(sys.argv[3], "r") as snapshot_taken:
        response = snapshot_taken.read()
    assert response == "done"
    print("[child] Continuing execution")

{body}
"""

ALLOCATE_THEN_FREE_THEN_SNAPSHOT = textwrap.dedent(
    f"""
        allocator = MemoryAllocator()

        with get_tracker():
            allocator.valloc({ALLOCATION_SIZE})
            allocator.free()
            snapshot_point()
    """
)
ALLOCATE_THEN_SNAPSHOT_THEN_FREE = textwrap.dedent(
    f"""
        allocator = MemoryAllocator()
        with get_tracker():
            allocator.valloc({ALLOCATION_SIZE})
            snapshot_point()
            allocator.free()
    """
)
ALLOCATE_MANY_THEN_SNAPSHOT_THEN_FREE_MANY = textwrap.dedent(
    f"""
        allocators = [MemoryAllocator() for _ in range({MULTI_ALLOCATION_COUNT})]
        with get_tracker():
            for allocator in allocators:
                allocator.valloc({ALLOCATION_SIZE})
            snapshot_point()
            for allocator in allocators:
                allocator.free()
    """
)


@contextmanager
def run_till_snapshot_point(
    program: str,
    *,
    reader: SocketReader,
    tmp_path: Path,
    free_port: int,
) -> Iterator[None]:
    allocations_made = tmp_path / "allocations_made.event"
    snapshot_taken = tmp_path / "snapshot_taken.event"
    os.mkfifo(allocations_made)
    os.mkfifo(snapshot_taken)

    script = _SCRIPT_TEMPLATE.format(body=program)

    env = os.environ.copy()
    env.pop("PYTHONMALLOC", None)
    proc = subprocess.Popen(
        [
            sys.executable,
            "-c",
            script,
            str(free_port),
            allocations_made,
            snapshot_taken,
        ],
        env=env,
    )

    try:
        with reader:
            print("[parent] Waiting on allocations made")
            with open(allocations_made, "r") as f1:
                assert f1.read() == "done"

            print("[parent] Deferring to caller")
            # Wait a bit of time, for background thread to receive + process the records.
            time.sleep(0.1)
            yield

            print("[parent] Notifying program to continue")
            with open(snapshot_taken, "w") as f2:
                f2.write("done")
            print("[parent] Will close socket reader now.")
    finally:
        print("[parent] Waiting on child to exit.")
        try:
            assert proc.wait(timeout=TIMEOUT) == 0
        except subprocess.TimeoutExpired:
            print("[parent] Killing child, after timeout.")
            proc.kill()
            raise


#
# Actual tests
#
class TestSocketReaderErrorHandling:
    @pytest.mark.valgrind
    def test_get_current_snapshot_raises_before_context(self, free_port: int) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)

        # WHEN / THEN
        with pytest.raises(StopIteration):
            next(reader.get_current_snapshot(merge_threads=False))

    def test_get_is_active_after_context(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_SNAPSHOT_THEN_FREE

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            pass

        # THEN
        assert reader.is_active is False

    @pytest.mark.valgrind
    def test_get_current_snapshot_raises_after_context(
        self, free_port: int, tmp_path: Path
    ) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_SNAPSHOT_THEN_FREE

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            pass

        # THEN
        with pytest.raises(StopIteration):
            next(reader.get_current_snapshot(merge_threads=False))

    @pytest.mark.valgrind
    def test_get_current_snapshot_first_yield_after_context_raises(
        self, free_port: int, tmp_path: Path
    ) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_SNAPSHOT_THEN_FREE

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            snapshot = reader.get_current_snapshot(merge_threads=False)

        # THEN
        with pytest.raises(StopIteration):
            next(snapshot)

    @pytest.mark.valgrind
    def test_nested_context_is_diallowed(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_FREE_THEN_SNAPSHOT

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            # THEN
            with pytest.raises(
                ValueError, match="Can not enter (.*)context (.*)more than once"
            ):
                with reader:
                    pass


class TestSocketReaderAccess:
    @pytest.mark.valgrind
    def test_empty_snapshot_after_free(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_FREE_THEN_SNAPSHOT

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            unfiltered_snapshot = list(reader.get_current_snapshot(merge_threads=False))

        # THEN
        snapshot = list(filter_relevant_allocations(unfiltered_snapshot))
        assert snapshot == []

    @pytest.mark.valgrind
    def test_single_allocation_snapshot(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_SNAPSHOT_THEN_FREE

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            unfiltered_snapshot = list(reader.get_current_snapshot(merge_threads=False))

        # THEN
        snapshot = list(filter_relevant_allocations(unfiltered_snapshot))
        assert len(snapshot) == 1

        allocation = snapshot[0]
        assert allocation.size == ALLOCATION_SIZE * 1
        assert allocation.allocator == AllocatorType.VALLOC

        symbol, filename, lineno = allocation.stack_trace()[0]
        assert symbol == "valloc"
        assert filename.endswith("/_test.py")
        assert 0 < lineno < 200

    @pytest.mark.valgrind
    def test_multi_allocation_snapshot(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_MANY_THEN_SNAPSHOT_THEN_FREE_MANY

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            unfiltered_snapshot = list(reader.get_current_snapshot(merge_threads=False))

        # THEN
        snapshot = list(filter_relevant_allocations(unfiltered_snapshot))
        assert len(snapshot) == 1

        allocation = snapshot[0]
        assert allocation.size == ALLOCATION_SIZE * 10
        assert allocation.allocator == AllocatorType.VALLOC

        symbol, filename, lineno = allocation.stack_trace()[0]
        assert symbol == "valloc"
        assert filename.endswith("/_test.py")
        assert 0 < lineno < 200

    @pytest.mark.valgrind
    def test_multiple_context_entries_does_not_crash(
        self, free_port: int, tmp_path: Path
    ) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_FREE_THEN_SNAPSHOT
        tmp_path_one = tmp_path / "one"
        tmp_path_one.mkdir()
        tmp_path_two = tmp_path / "two"
        tmp_path_two.mkdir()

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path_one,
            free_port=free_port,
        ):
            pass

        # THEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path_two,
            free_port=free_port,
        ):
            pass

    @pytest.mark.valgrind
    def test_command_line(self, free_port: int, tmp_path: Path) -> None:
        # GIVEN
        reader = SocketReader(port=free_port)
        program = ALLOCATE_THEN_FREE_THEN_SNAPSHOT

        # WHEN
        with run_till_snapshot_point(
            program,
            reader=reader,
            tmp_path=tmp_path,
            free_port=free_port,
        ):
            command_line = reader.command_line

        # THEN
        assert command_line
        assert command_line.startswith("-c")  # these samples run with `python -c`
        assert str(free_port) in command_line

    @pytest.mark.valgrind
    def test_reading_allocations_while_reading_stack_traces(
        self, free_port: int, tmp_path: Path
    ) -> None:
        """This test exist mainly to give valgrind/helgrind the opportunity to
        see races between fetching stack traces and the background thread
        fetching allocations. Therefore no interesting asserts are done in
        the test itself.

        The test will spawn a tracked process that is constantly creating new
        functions and code objects so the process is constantly sending new
        frames to the SocketReader, triggering tons of updates of the internal
        structures inside. We want to check that if we get the stack traces
        inside the allocations it won't crash/race with fetching new allocations
        """
        # GIVEN
        script = textwrap.dedent(
            f"""\
            from memray._test import MemoryAllocator
            from memray._memray import SocketDestination
            from memray._memray import Tracker
            from itertools import count
            import textwrap

            def foo(allocator, n):
                fname = f"blech_{{n}}"
                namespace = {{}}

                exec(textwrap.dedent(f'''
                def {{fname}}(allocator):
                    allocator.valloc(1024)
                    allocator.free()
                '''), namespace, namespace)

                func = namespace[fname]
                func(allocator)

            with Tracker(destination=SocketDestination(server_port={free_port})):
                allocator = MemoryAllocator()
                for n in count(0):
                    foo(allocator, n)
        """
        )
        code = [
            sys.executable,
            "-c",
            script,
        ]
        traces: Set[Tuple[str, str, int]] = set()
        MAX_TRACES = 10

        # WHEN
        with subprocess.Popen(code) as proc:
            with SocketReader(port=free_port) as reader:
                while len(traces) < MAX_TRACES:
                    for allocation in reader.get_current_snapshot(merge_threads=False):
                        traces.update(allocation.stack_trace())

                proc.terminate()

        # THEN
        assert len(traces) >= MAX_TRACES



import numpy as np

# DO NOT CHANGE
MB_CONVERSION = 1000 * 1000
DUPLICATE_CONST = 2

SIZE_OF_DATA_IN_MB = 100
SUBTRACT_AMOUNT = 3
POWER_AMOUNT = 2
ADD_AMOUNT = 10
# DO NOT CHANGE


def load_xMb_of_data(mb: int) -> np.ndarray:
    size = MB_CONVERSION * mb  # DO NOT CHANGE
    return np.ones(size, dtype=np.uint8)


def duplicate_data(data: np.ndarray) -> np.ndarray:
    return data * DUPLICATE_CONST


def add_scalar(data: np.ndarray, scalar: int) -> np.ndarray:
    return data + scalar


def subtract_scalar(data: np.ndarray, scalar: int) -> np.ndarray:
    return data - scalar


def raise_to_power(data: np.ndarray, power: int) -> np.ndarray:
    return np.power(data, power)


def process_data() -> np.ndarray:
    data = load_xMb_of_data(SIZE_OF_DATA_IN_MB)
    data = subtract_scalar(data, SUBTRACT_AMOUNT)
    data_pow = raise_to_power(data, POWER_AMOUNT)
    return add_scalar(duplicate_data(data_pow), ADD_AMOUNT)


if __name__ == "__main__":
    process_data()



import shutil
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

import pytest

from memray import AllocatorType
from memray import FileReader
from memray import Tracker
from memray._test import MemoryAllocator

HERE = Path(__file__).parent
TEST_MULTITHREADED_EXTENSION = HERE / "multithreaded_extension"
TEST_MISBEHAVING_EXTENSION = HERE / "misbehaving_extension"
TEST_RPATH_EXTENSION = HERE / "rpath_extension"


@pytest.mark.valgrind
def test_multithreaded_extension(tmpdir, monkeypatch):
    """Test tracking allocations in a native extension which spawns multiple threads,
    each thread allocating and freeing memory."""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "multithreaded_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MULTITHREADED_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from testext import run  # type: ignore

        with Tracker(output):
            run()

    # THEN
    records = list(FileReader(output).get_allocation_records())
    assert records

    memaligns = [
        record for record in records if record.allocator == AllocatorType.POSIX_MEMALIGN
    ]
    assert len(memaligns) == 100 * 100  # 100 threads allocate 100 times in testext

    # We don't keep track of the native stacks. Make sure they are empty
    assert all(len(memalign.stack_trace()) == 0 for memalign in memaligns)

    memaligns_addr = {record.address for record in memaligns}
    memalign_frees = [
        record
        for record in records
        if record.address in memaligns_addr and record.allocator == AllocatorType.FREE
    ]

    assert len(memalign_frees) >= 100 * 100


def test_misbehaving_extension(tmpdir, monkeypatch):
    """Check that we can correctly track allocations in an extension which invokes
    Python code in a thread and does not register trace functions."""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "misbehaving_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MISBEHAVING_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    def allocating_function():  # pragma: no cover
        allocator = MemoryAllocator()
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from misbehaving import call_fn  # type: ignore

        with Tracker(output):
            call_fn(allocating_function)

    # THEN
    allocations = list(FileReader(output).get_allocation_records())
    allocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(allocs) == 1
    (alloc,) = allocs

    stack_trace = alloc.stack_trace()
    assert len(stack_trace)

    *_, bottom_frame = stack_trace
    func, filename, line = bottom_frame
    assert func == "allocating_function"
    assert filename.endswith(__file__)
    assert line == 83

    frees = [
        event
        for event in allocations
        if event.address == alloc.address and event.allocator == AllocatorType.FREE
    ]
    assert len(frees) >= 1


def test_extension_that_uses_pygilstate_ensure(tmpdir, monkeypatch):
    """Check that we can correctly track allocations in an extension which invokes
    Python code in a thread and does not register trace functions."""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "misbehaving_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MISBEHAVING_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    def allocating_function():
        allocator = MemoryAllocator()
        allocator.valloc(1234)
        allocator.free()

    def foo1():
        foo2()

    def foo2():
        call_fn_no_thread(allocating_function)

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from misbehaving import call_fn_no_thread

        allocator = MemoryAllocator()
        with Tracker(output):
            foo1()
            allocator.valloc(1234)
            allocator.free()

    # THEN
    allocations = list(FileReader(output).get_allocation_records())
    allocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(allocs) == 2
    (alloc1, alloc2) = allocs

    stack_trace = alloc1.stack_trace()
    assert len(stack_trace)
    first_frame, *_, bottom_frame = stack_trace
    func, filename, line = bottom_frame
    assert func == "test_extension_that_uses_pygilstate_ensure"
    assert filename.endswith(__file__)
    assert line == 154

    # We should have 2 frames here: this function calling `allocator.valloc`,
    # and `allocator.valloc` calling the C `valloc`.
    # We should not see any call to foo1() or foo2().
    stack_trace = alloc2.stack_trace()
    assert len(stack_trace) == 2
    (callee, caller) = stack_trace
    func, filename, line = callee
    assert func == "valloc"
    assert filename.endswith("/_test.py")

    func, filename, line = caller
    assert func == "test_extension_that_uses_pygilstate_ensure"
    assert filename.endswith(__file__)
    assert line == 155

    frees = [
        event
        for event in allocations
        if event.address in (alloc1.address, alloc2.address)
        and event.allocator == AllocatorType.FREE
    ]
    assert len(frees) >= 1


def test_native_dlopen(tmpdir, monkeypatch):
    """Check that we can correctly track allocations in an extension which calls
    dlopen() without the GIL held"""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "misbehaving_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MISBEHAVING_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    def allocating_function():
        allocator = MemoryAllocator()
        allocator.valloc(1234)
        allocator.free()

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from misbehaving import dlopen_self  # type: ignore

        with Tracker(output):
            dlopen_self(allocating_function)

    # THEN
    allocations = list(FileReader(output).get_allocation_records())
    allocs = [
        event
        for event in allocations
        if event.size == 1234 and event.allocator == AllocatorType.VALLOC
    ]
    assert len(allocs) == 1
    (alloc,) = allocs

    stack_trace = alloc.stack_trace()
    assert len(stack_trace)

    *_, bottom_frame = stack_trace
    func, filename, line = bottom_frame
    assert func == "test_native_dlopen"
    assert filename.endswith(__file__)
    assert line == 226

    frees = [
        event
        for event in allocations
        if event.address == alloc.address and event.allocator == AllocatorType.FREE
    ]
    assert len(frees) >= 1


@pytest.mark.valgrind
def test_valloc_at_thread_exit(tmpdir, monkeypatch):
    """Test tracking allocations that happen while a thread is shutting down"""
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "multithreaded_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MULTITHREADED_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from testext import run_valloc_at_exit  # type: ignore

        with Tracker(output, native_traces=True):
            run_valloc_at_exit()

    # THEN
    records = list(FileReader(output).get_allocation_records())
    assert records

    vallocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(vallocs) == 1


def test_valloc_at_thread_exit_in_subprocess(tmpdir, monkeypatch):
    """Test tracking allocations in the destructor of a TLS variable.

    Ensure that TLS variable is created before Memray is imported.
    """
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "multithreaded_extension"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_MULTITHREADED_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    code = dedent(
        f"""
        from testext import run_valloc_at_exit
        run_valloc_at_exit()  # First call creates the test extension TLS.

        from memray import Tracker
        with Tracker({str(output)!r}, native_traces=True):
            run_valloc_at_exit()
        """
    )

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setenv("PYTHONPATH", str(extension_path), prepend=":")
        subprocess.run(
            [sys.executable, "-c", code],
            check=True,
        )

    # THEN
    records = list(FileReader(output).get_allocation_records())
    assert records

    vallocs = [record for record in records if record.allocator == AllocatorType.VALLOC]
    assert len(vallocs) == 1


@pytest.mark.parametrize("py_finalize", [True, False])
def test_hard_exit(tmpdir, py_finalize):
    """Test a program that exits directly under the context manager"""

    # GIVEN

    # Run a program that calls from memray._test.exit under a Tracker context
    # manager and check that it finishes without error

    output = Path(tmpdir) / "test.bin"
    code = dedent(
        f"""\
    from memray._test import exit
    from memray import Tracker
    with Tracker("{output}"):
        exit(py_finalize={py_finalize})
    """
    )

    # WHEN
    subprocess.run(
        [sys.executable, "-c", code],
        check=True,
        capture_output=True,
    )

    # THEN
    # No assertions, just check that the program exits without error


@pytest.mark.skipif(
    sys.platform == "darwin", reason="Test requires a linker that supports $ORIGIN"
)
def test_dlopen_with_rpath(tmpdir, monkeypatch):
    # GIVEN
    output = Path(tmpdir) / "test.bin"
    extension_name = "sharedlibs"
    extension_path = tmpdir / extension_name
    shutil.copytree(TEST_RPATH_EXTENSION, extension_path)
    subprocess.run(
        [sys.executable, str(extension_path / "setup.py"), "build_ext", "--inplace"],
        check=True,
        cwd=extension_path,
        capture_output=True,
    )

    # WHEN
    with monkeypatch.context() as ctx:
        ctx.setattr(sys, "path", [*sys.path, str(extension_path)])
        from ext import hello_world  # type: ignore

        try:
            hello_world()
        except RuntimeError:
            pytest.skip("Test requires a linker that supports -rpath")

        # THEN
        with Tracker(output):
            hello_world()



import functools
import subprocess
import sys

import pytest

from memray import AllocatorType
from memray import FileReader
from memray.commands.attach import debugger_available
from tests.utils import filter_relevant_allocations

PROGRAM = """
import sys
import threading
import time

from memray._test import MemoryAllocator

stop_looping = threading.Event()


def io_thread_body():
    print("ready")
    for line in sys.stdin:
        stop_looping.set()


def alloc_thread_body():
    while not stop_looping.is_set():
        allocator = MemoryAllocator()
        allocator.malloc(1024)
        allocator.free()


def foo():
    bar()


def bar():
    baz()


def baz():
    allocator = MemoryAllocator()
    allocator.valloc(50 * 1024 * 1024)
    allocator.free()


# Debugger attaches wait for allocator or deallocator calls, so we can't just
# block waiting for a signal. Allocate and free in a loop in the background.
# Additionally, there must be a Python loop in the main thread, because that's
# the thread `sys.remote_exec` attaches to and it needs the eval breaker. Spawn
# another background thread to let us know when to stop looping.
io_thread = threading.Thread(target=io_thread_body)
io_thread.start()

alloc_thread = threading.Thread(target=alloc_thread_body)
alloc_thread.start()

while not stop_looping.is_set():
    time.sleep(0.1)

alloc_thread.join()
io_thread.join()

foo()
"""


def generate_attach_command(method, output, *args):
    cmd = [
        sys.executable,
        "-m",
        "memray",
        "attach",
        "--verbose",
        "--force",
        "--method",
        method,
        "-o",
        str(output),
    ]

    if args:
        cmd.extend(args)

    return cmd


def generate_detach_command(method, *args):
    cmd = [
        sys.executable,
        "-m",
        "memray",
        "detach",
        "--verbose",
        "--method",
        method,
    ]

    if args:
        cmd.extend(args)

    return cmd


def run_process(cmd, wait_for_stderr=False):
    process_stderr = ""
    tracked_process = subprocess.Popen(
        [sys.executable, "-Wignore", "-uc", PROGRAM],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )

    # Help the type checker out...
    assert tracked_process.stdin is not None
    assert tracked_process.stdout is not None

    assert tracked_process.stdout.readline() == "ready\n"

    cmd.append(str(tracked_process.pid))

    # WHEN
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, text=True)
    except subprocess.CalledProcessError as exc:
        # The test has failed; we'd just wait forever.
        wait_for_stderr = False

        if "Couldn't write extended state status" in exc.output:
            # https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=898048
            pytest.xfail("gdb < 8 does not support this CPU")
        else:
            print(exc.output)
            raise
    finally:
        print("1", file=tracked_process.stdin, flush=True)
        if wait_for_stderr:
            process_stderr = tracked_process.stderr.readline()
            # Skip any lines like 'WARNING: Correcting symbol for malloc'
            while "warning" in process_stderr.lower():
                process_stderr = tracked_process.stderr.readline()
        tracked_process.stdin.close()
        tracked_process.wait()

    # THEN
    assert "" == tracked_process.stdout.read()
    assert tracked_process.returncode == 0
    return process_stderr


def get_call_stack(allocation):
    return [f[0] for f in allocation.stack_trace()]


def get_relevant_vallocs(records):
    return [
        record
        for record in filter_relevant_allocations(records)
        if record.allocator == AllocatorType.VALLOC
    ]


@functools.lru_cache(maxsize=None)
def debugging_method_works(method):
    proc = subprocess.Popen(
        [sys.executable, "-Wignore", "-uc", r'input("ready\n")'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        text=True,
    )
    assert proc.stdout.readline() == "ready\n"
    try:
        if method == "sys.remote_exec":
            cmd = [
                sys.executable,
                "-uc",
                f"import sys; sys.remote_exec({proc.pid}, '/dev/null')",
            ]
        elif method == "gdb":
            cmd = ["gdb", "-batch", "-p", str(proc.pid)]
        elif method == "lldb":
            cmd = ["lldb", "--batch", "-p", str(proc.pid)]
        else:
            raise ValueError(f"Unknown method: {method}")

        try:
            subprocess.check_call(cmd)
            return True
        except Exception:
            return False
    finally:
        proc.terminate()
        proc.wait()


def skip_if_not_supported(method):
    if not debugger_available(method):
        pytest.skip(f"{method} isn't available")

    if not debugging_method_works(method):
        pytest.skip(f"{method} cannot attach to remote processes on this system")


@pytest.mark.parametrize("method", ["sys.remote_exec", "lldb", "gdb"])
def test_basic_attach(tmp_path, method):
    skip_if_not_supported(method)

    # GIVEN
    output = tmp_path / "test.bin"
    attach_cmd = generate_attach_command(method, output)

    # WHEN
    run_process(attach_cmd)

    # THEN
    reader = FileReader(output)
    (valloc,) = get_relevant_vallocs(reader.get_allocation_records())
    assert get_call_stack(valloc) == ["valloc", "baz", "bar", "foo", "<module>"]


@pytest.mark.parametrize("method", ["sys.remote_exec", "lldb", "gdb"])
def test_aggregated_attach(tmp_path, method):
    skip_if_not_supported(method)

    # GIVEN
    output = tmp_path / "test.bin"
    attach_cmd = generate_attach_command(method, output, "--aggregate")

    # WHEN
    run_process(attach_cmd)

    # THEN
    reader = FileReader(output)
    with pytest.raises(
        NotImplementedError,
        match="Can't get all allocations from a pre-aggregated capture file.",
    ):
        list(reader.get_allocation_records())

    (valloc,) = get_relevant_vallocs(reader.get_high_watermark_allocation_records())
    assert get_call_stack(valloc) == ["valloc", "baz", "bar", "foo", "<module>"]


@pytest.mark.parametrize("method", ["sys.remote_exec", "lldb", "gdb"])
def test_attach_time(tmp_path, method):
    skip_if_not_supported(method)

    # GIVEN
    output = tmp_path / "test.bin"
    attach_cmd = generate_attach_command(method, output, "--duration", "1")

    # WHEN
    process_stderr = run_process(attach_cmd, wait_for_stderr=True)

    # THEN
    assert "memray: Deactivating tracking: 1 seconds have elapsed" in process_stderr


@pytest.mark.parametrize("method", ["sys.remote_exec", "lldb", "gdb"])
def test_detach_without_attach(method):
    skip_if_not_supported(method)

    # GIVEN
    detach_cmd = generate_detach_command(method)

    # WHEN
    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        run_process(detach_cmd)

    # THEN
    assert (
        "Failed to stop tracking in remote process:"
        " no previous `memray attach` call detected"
    ) in exc_info.value.stdout



"""
In this example we investigate the primordial Earth embedded in a disk of
planetesimals, integrating it for a short period of time using the MERCURIUS
integrator. MERCURIUS is a hybrid integration scheme which combines the WHFAST
and IAS15 algorithms in a similar way to the hybrid integrastor in the MERCURY
package.
"""

import numpy as np
import rebound

# First let's choose the basic properties required for the MERCURIUS
# integrator. In particular, we are * setting planetesimals to semi-active
# mode, which means they can influence active bodies but not other semi-active
# bodies. * merge any planetesimals that collide with a planets, conserving
# momentum and mass. * remove particles from the similation which leave our
# pre-defined box. * track the energy lost due to ejections or collisions.
sim = rebound.Simulation()

# integrator options
sim.integrator = "mercurius"
sim.dt = 0.025 * 2.0 * np.pi  # we're working in units where 1 year = 2*pi
sim.testparticle_type = 1
sim.ri_ias15.min_dt = 1e-6  # ensure that close encounters do not stall the integration

# collision and boundary options
sim.collision = "direct"
sim.collision_resolve = "merge"
sim.collision_resolve_keep_sorted = 1
sim.track_energy_offset = 1

# Now that the setup is complete, it's time to add some particles! When using
# the MERCURIUS integrator it is important to add active bodies first and
# semi-active bodies later. The N_active variable separates massive bodies from
# semi-active/test bodies. Here, we add two active particles, the Sun and the
# Earth. Thus, N_active will be 2.

sim.add(m=1.0)
sim.add(m=3e-6, r=5e-5, a=1, e=0.05, f=np.pi)
sim.N_active = sim.N  # sim.N= 2

# Now, let's create our planetesimal disk. First we define three different
# distribution functions - powerlaw, uniform and rayleigh.


def rand_powerlaw(slope, min_v, max_v):
    y = np.random.uniform()
    pow_max = pow(max_v, slope + 1.0)
    pow_min = pow(min_v, slope + 1.0)
    return pow((pow_max - pow_min) * y + pow_min, 1.0 / (slope + 1.0))


def rand_uniform(minimum, maximum):
    return np.random.uniform() * (maximum - minimum) + minimum


def rand_rayleigh(sigma):
    return sigma * np.sqrt(-2 * np.log(np.random.uniform()))


# Next, let's set up the basic properties of our planetesimal disk. For this
# simple example we are assuming that all planetesimals have the same mass and
# radius.

N_pl = 50000  # Number of planetesimals
Mtot_disk = 10 * sim.particles[1].m  # Total mass of planetesimal disk
m_pl = Mtot_disk / float(N_pl)  # Mass of each planetesimal
r_pl = 2e-5  # Radius of each planetesimal

# Now let's add our planetesimals to the simulation!

np.random.seed(42)  # by setting a seed we will reproduce the same simulation every time
while sim.N < (N_pl + sim.N_active):
    a = rand_powerlaw(0, 0.99, 1.01)
    e = rand_rayleigh(0.01)
    inc = rand_rayleigh(0.005)
    f = rand_uniform(-np.pi, np.pi)
    p = rebound.Particle(
        simulation=sim,
        primary=sim.particles[0],
        m=m_pl,
        r=r_pl,
        a=a,
        e=e,
        inc=inc,
        Omega=0,
        omega=0,
        f=f,
    )
    # Only add planetesimal if it's far away from the planet
    d = np.linalg.norm(np.array(p.xyz) - np.array(sim.particles[1].xyz))
    if d > 0.01:
        sim.add(p)

# We move to the COM frame to avoid having the simulation drive away from the
# origin. In addition, it is always good practice to monitor the change in
# energy over the course of a simulation, which requires us to calculate it
# before and after the simulation.

sim.move_to_com()
E0 = sim.calculate_energy()

# Finally, let us simulate our system for 1 year, and check that our final
# relative energy error is small.

times = np.linspace(0.0, 10.0, 10)
encounterN = np.zeros(len(times))
totalN = np.zeros(len(times))
errors = np.zeros(len(times))
for i, t in enumerate(times):
    sim.integrate(t, exact_finish_time=0)
    totalN[i] = sim.N
    encounterN[i] = sim.ri_mercurius._encounterN
    errors[i] = abs((sim.calculate_energy() - E0) / E0)



from unittest.mock import patch

import pytest

from memray.commands import main


@patch("memray.commands.attach.debugger_available")
class TestAttachSubCommand:
    def test_memray_attach_aggregated_without_output_file(
        self, is_debugger_available_mock, capsys
    ):
        # GIVEN
        is_debugger_available_mock.return_value = True

        # WHEN
        with pytest.raises(SystemExit):
            main(["attach", "--aggregate", "1234"])

        captured = capsys.readouterr()
        print("Error", captured.err)
        assert "Can't use aggregated mode without an output file." in captured.err
